"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var CompressedJSON_1 = require("./CompressedJSON");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var MIN_LENGTH_FOR_ENUM = 10;
function concatArrays(arrays) {
    var combined = [];
    for (var i = 0; i < arrays.length; i++)
        combined = combined.concat(arrays[i]);
    return combined;
}
var InferenceUnionBuilder = /** @class */ (function (_super) {
    __extends(InferenceUnionBuilder, _super);
    function InferenceUnionBuilder(typeBuilder, typeName, _typeInference, _cjson, _numValues) {
        var _this = _super.call(this, typeBuilder, typeName, true) || this;
        _this._typeInference = _typeInference;
        _this._cjson = _cjson;
        _this._numValues = _numValues;
        return _this;
    }
    InferenceUnionBuilder.prototype.makeEnum = function (enumCases) {
        Support_1.assert(enumCases.length > 0);
        if (enumCases.length < Math.sqrt(this._numValues)) {
            return this.typeBuilder.getEnumType(this.typeName, true, immutable_1.OrderedSet(enumCases));
        }
        return null;
    };
    InferenceUnionBuilder.prototype.makeClass = function (classes, maps) {
        Support_1.assert(maps.length === 0);
        return this._typeInference.inferClassType(this._cjson, this.typeName, classes);
    };
    InferenceUnionBuilder.prototype.makeArray = function (arrays) {
        var combined = concatArrays(arrays);
        return this.typeBuilder.getArrayType(this._typeInference.inferType(this._cjson, pluralize.singular(this.typeName), combined));
    };
    return InferenceUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
var TypeInference = /** @class */ (function () {
    function TypeInference(_inferMaps, _inferEnums) {
        var _this = this;
        this._inferMaps = _inferMaps;
        this._inferEnums = _inferEnums;
        this._typeBuilder = new TypeBuilder_1.TypeBuilder();
        this.inferType = function (cjson, typeName, valueArray) {
            var unionBuilder = new InferenceUnionBuilder(_this._typeBuilder, typeName, _this, cjson, valueArray.length);
            for (var _i = 0, valueArray_1 = valueArray; _i < valueArray_1.length; _i++) {
                var value = valueArray_1[_i];
                var t = CompressedJSON_1.valueTag(value);
                switch (t) {
                    case CompressedJSON_1.Tag.Null:
                        unionBuilder.addNull();
                        break;
                    case CompressedJSON_1.Tag.False:
                    case CompressedJSON_1.Tag.True:
                        unionBuilder.addBool();
                        break;
                    case CompressedJSON_1.Tag.Integer:
                        unionBuilder.addInteger();
                        break;
                    case CompressedJSON_1.Tag.Double:
                        unionBuilder.addDouble();
                        break;
                    case CompressedJSON_1.Tag.InternedString:
                        if (_this._inferEnums && !unionBuilder.haveString && valueArray.length >= MIN_LENGTH_FOR_ENUM) {
                            var s = cjson.getStringForValue(value);
                            unionBuilder.addEnumCase(s);
                        }
                        else {
                            unionBuilder.addString();
                        }
                        break;
                    case CompressedJSON_1.Tag.UninternedString:
                        unionBuilder.addString();
                        break;
                    case CompressedJSON_1.Tag.Object:
                        unionBuilder.addClass(cjson.getObjectForValue(value));
                        break;
                    case CompressedJSON_1.Tag.Array:
                        unionBuilder.addArray(cjson.getArrayForValue(value));
                        break;
                    default:
                        return Support_1.assertNever(t);
                }
            }
            var result = unionBuilder.buildUnion(false);
            if (result.isNamedType()) {
                result.setGivenName(typeName);
            }
            return result;
        };
        this.inferClassType = function (cjson, typeName, objects) {
            var combined = concatArrays(objects);
            var propertyNames = [];
            var propertyValues = {};
            for (var i = 0; i < combined.length; i += 2) {
                var key = cjson.getStringForValue(combined[i]);
                var value = combined[i + 1];
                if (!Object.prototype.hasOwnProperty.call(propertyValues, key)) {
                    propertyNames.push(key);
                    propertyValues[key] = [];
                }
                propertyValues[key].push(value);
            }
            var properties = [];
            var couldBeMap = _this._inferMaps;
            for (var _i = 0, propertyNames_1 = propertyNames; _i < propertyNames_1.length; _i++) {
                var key = propertyNames_1[_i];
                var values = propertyValues[key];
                var t = _this.inferType(cjson, key, values);
                if (values.length < objects.length) {
                    t = Type_1.makeNullable(t, key, true);
                }
                if (couldBeMap && properties.length > 0 && t !== properties[0][1]) {
                    couldBeMap = false;
                }
                properties.push([key, t]);
            }
            if (couldBeMap && properties.length >= 20) {
                return _this._typeBuilder.getMapType(properties[0][1]);
            }
            return _this._typeBuilder.getClassType(typeName, true, immutable_1.OrderedMap(properties));
        };
    }
    return TypeInference;
}());
exports.TypeInference = TypeInference;
