"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var stringHash = require("string-hash");
var Support_1 = require("./Support");
var Type = /** @class */ (function () {
    function Type(kind) {
        this.kind = kind;
    }
    Type.prototype.isNamedType = function () {
        return false;
    };
    Type.prototype.directlyReachableTypes = function (setForType) {
        var set = setForType(this);
        if (set)
            return set;
        return orderedSetUnion(this.children.map(function (t) { return t.directlyReachableTypes(setForType); }));
    };
    Type.prototype.equals = function (other) {
        return typesEqual(this, other);
    };
    Type.prototype.hashCode = function () {
        return this.flatHashCode();
    };
    return Type;
}());
exports.Type = Type;
var PrimitiveType = /** @class */ (function (_super) {
    __extends(PrimitiveType, _super);
    function PrimitiveType(kind) {
        return _super.call(this, kind) || this;
    }
    Object.defineProperty(PrimitiveType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(PrimitiveType.prototype, "isNullable", {
        get: function () {
            return this.kind === "null";
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.map = function (f) {
        return this;
    };
    PrimitiveType.prototype.expandForEquality = function (other) {
        if (!(other instanceof PrimitiveType))
            return false;
        return this.kind === other.kind;
    };
    PrimitiveType.prototype.flatHashCode = function () {
        return stringHash(this.kind) | 0;
    };
    return PrimitiveType;
}(Type));
exports.PrimitiveType = PrimitiveType;
function isNull(t) {
    return t.kind === "null";
}
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    function ArrayType(items) {
        var _this = _super.call(this, "array") || this;
        _this.items = items;
        return _this;
    }
    Object.defineProperty(ArrayType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.items]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ArrayType.prototype.map = function (f) {
        var items = f(this.items);
        if (items === this.items)
            return this;
        return new ArrayType(items);
    };
    ArrayType.prototype.expandForEquality = function (other) {
        if (!(other instanceof ArrayType))
            return false;
        return [[this.items, other.items]];
    };
    ArrayType.prototype.flatHashCode = function () {
        return (stringHash(this.kind) + this.items.hashCode()) | 0;
    };
    return ArrayType;
}(Type));
exports.ArrayType = ArrayType;
var MapType = /** @class */ (function (_super) {
    __extends(MapType, _super);
    function MapType(values) {
        var _this = _super.call(this, "map") || this;
        _this.values = values;
        return _this;
    }
    Object.defineProperty(MapType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.values]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    MapType.prototype.map = function (f) {
        var values = f(this.values);
        if (values === this.values)
            return this;
        return new MapType(values);
    };
    MapType.prototype.expandForEquality = function (other) {
        if (!(other instanceof MapType))
            return false;
        return [[this.values, other.values]];
    };
    MapType.prototype.flatHashCode = function () {
        return (stringHash(this.kind) + this.values.hashCode()) | 0;
    };
    return MapType;
}(Type));
exports.MapType = MapType;
// FIXME: In the case of overlapping prefixes and suffixes we will
// produce a name that includes the overlap twice.  For example, for
// the names "aaa" and "aaaa" we have the common prefix "aaa" and the
// common suffix "aaa", so we will produce the combined name "aaaaaa".
function combineNames(names) {
    var first = names.first();
    if (first === undefined) {
        return Support_1.panic("Named type has no names");
    }
    if (names.count() === 1) {
        return first;
    }
    var prefixLength = first.length;
    var suffixLength = first.length;
    names.rest().forEach(function (n) {
        prefixLength = Math.min(prefixLength, n.length);
        for (var i = 0; i < prefixLength; i++) {
            if (first[i] !== n[i]) {
                prefixLength = i;
                break;
            }
        }
        suffixLength = Math.min(suffixLength, n.length);
        for (var i = 0; i < suffixLength; i++) {
            if (first[first.length - i - 1] !== n[n.length - i - 1]) {
                suffixLength = i;
                break;
            }
        }
    });
    var prefix = prefixLength > 2 ? first.substr(0, prefixLength) : "";
    var suffix = suffixLength > 2 ? first.substr(first.length - suffixLength) : "";
    var combined = prefix + suffix;
    if (combined.length > 2) {
        return combined;
    }
    return first;
}
var NamedType = /** @class */ (function (_super) {
    __extends(NamedType, _super);
    function NamedType(kind, nameOrNames, areNamesInferred) {
        var _this = _super.call(this, kind) || this;
        if (typeof nameOrNames === "string") {
            _this._names = immutable_1.OrderedSet([nameOrNames]);
        }
        else {
            _this._names = nameOrNames;
        }
        _this._areNamesInferred = areNamesInferred;
        return _this;
    }
    NamedType.prototype.isNamedType = function () {
        return true;
    };
    Object.defineProperty(NamedType.prototype, "names", {
        get: function () {
            return this._names;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NamedType.prototype, "areNamesInferred", {
        get: function () {
            return this._areNamesInferred;
        },
        enumerable: true,
        configurable: true
    });
    NamedType.prototype.addName = function (name, isInferred) {
        if (isInferred && !this._areNamesInferred) {
            return;
        }
        if (this._areNamesInferred && !isInferred) {
            this._names = immutable_1.OrderedSet([name]);
            this._areNamesInferred = isInferred;
        }
        else {
            this._names = this._names.add(name);
        }
    };
    NamedType.prototype.setGivenName = function (name) {
        this._names = immutable_1.OrderedSet([name]);
        this._areNamesInferred = false;
    };
    Object.defineProperty(NamedType.prototype, "combinedName", {
        get: function () {
            return combineNames(this._names);
        },
        enumerable: true,
        configurable: true
    });
    return NamedType;
}(Type));
exports.NamedType = NamedType;
var ClassType = /** @class */ (function (_super) {
    __extends(ClassType, _super);
    function ClassType(names, areNamesInferred, _properties) {
        var _this = _super.call(this, "class", names, areNamesInferred) || this;
        _this._properties = _properties;
        return _this;
    }
    ClassType.prototype.setProperties = function (properties) {
        if (this._properties !== undefined) {
            return Support_1.panic("Can only set class properties once");
        }
        this._properties = properties;
    };
    Object.defineProperty(ClassType.prototype, "properties", {
        get: function () {
            if (this._properties === undefined) {
                return Support_1.panic("Class properties accessed before they were set");
            }
            return this._properties;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "sortedProperties", {
        get: function () {
            var _this = this;
            var sortedKeys = this.properties.keySeq().sort();
            var props = sortedKeys.map(function (k) { return [k, Support_1.defined(_this.properties.get(k))]; });
            return immutable_1.OrderedMap(props);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "children", {
        get: function () {
            return this.sortedProperties.toOrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ClassType.prototype.map = function (f) {
        var same = true;
        var properties = this.properties.map(function (t) {
            var ft = f(t);
            if (ft !== t)
                same = false;
            return ft;
        });
        if (same)
            return this;
        return new ClassType(this.names, this.areNamesInferred, properties);
    };
    ClassType.prototype.expandForEquality = function (other) {
        if (!(other instanceof ClassType))
            return false;
        if (!this.names.equals(other.names))
            return false;
        if (this.properties.size !== other.properties.size)
            return false;
        if (this.properties.size === 0)
            return true;
        var queue = [];
        this.properties.forEach(function (t, name) {
            var otherT = other.properties.get(name);
            if (!otherT)
                return false;
            queue.push([t, otherT]);
        });
        if (queue.length !== this.properties.size)
            return false;
        return queue;
    };
    ClassType.prototype.flatHashCode = function () {
        return (stringHash(this.kind) + this.names.hashCode() + this.properties.size) | 0;
    };
    ClassType.prototype.hashCode = function () {
        var hash = this.flatHashCode();
        this.properties.forEach(function (t, n) {
            hash = (hash + t.flatHashCode() + stringHash(n)) | 0;
        });
        return hash;
    };
    return ClassType;
}(NamedType));
exports.ClassType = ClassType;
var EnumType = /** @class */ (function (_super) {
    __extends(EnumType, _super);
    function EnumType(names, areNamesInferred, cases) {
        var _this = _super.call(this, "enum", names, areNamesInferred) || this;
        _this.cases = cases;
        return _this;
    }
    Object.defineProperty(EnumType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    EnumType.prototype.map = function (f) {
        return this;
    };
    EnumType.prototype.expandForEquality = function (other) {
        if (!(other instanceof EnumType))
            return false;
        return this.names.equals(other.names) && this.cases.equals(other.cases);
    };
    EnumType.prototype.flatHashCode = function () {
        return (stringHash(this.kind) + this.names.hashCode() + this.cases.hashCode()) | 0;
    };
    return EnumType;
}(NamedType));
exports.EnumType = EnumType;
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    function UnionType(names, areNamesInferred, members) {
        var _this = _super.call(this, "union", names, areNamesInferred) || this;
        _this.members = members;
        _this.findMember = function (kind) {
            return _this.members.find(function (t) { return t.kind === kind; });
        };
        Support_1.assert(members.size > 1);
        return _this;
    }
    Object.defineProperty(UnionType.prototype, "children", {
        get: function () {
            return this.sortedMembers;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "isNullable", {
        get: function () {
            return this.findMember("null") !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.map = function (f) {
        var same = true;
        var members = this.members.map(function (t) {
            var ft = f(t);
            if (ft !== t)
                same = false;
            return ft;
        });
        if (same)
            return this;
        return new UnionType(this.names, this.areNamesInferred, members);
    };
    Object.defineProperty(UnionType.prototype, "sortedMembers", {
        get: function () {
            // FIXME: We're assuming no two members of the same kind.
            return this.members.sortBy(function (t) { return t.kind; });
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.equals = function (other) {
        if (!(other instanceof UnionType))
            return false;
        return this.names.equals(other.names) && this.members.equals(other.members);
    };
    UnionType.prototype.expandForEquality = function (other) {
        if (!(other instanceof UnionType))
            return false;
        if (!this.names.equals(other.names))
            return false;
        if (this.members.size !== other.members.size)
            return false;
        if (this.members.size === 0)
            return true;
        var otherByKind = {};
        other.members.forEach(function (t) {
            otherByKind[t.kind] = t;
        });
        var queue = [];
        this.members.forEach(function (t) {
            var otherT = otherByKind[t.kind];
            if (!otherT)
                return false;
            queue.push([t, otherT]);
        });
        if (queue.length !== this.members.size)
            return false;
        return queue;
    };
    UnionType.prototype.flatHashCode = function () {
        return (stringHash(this.kind) + this.names.hashCode() + this.members.size) | 0;
    };
    UnionType.prototype.hashCode = function () {
        var hash = this.flatHashCode();
        this.members.forEach(function (t) {
            hash = (hash + t.flatHashCode()) | 0;
        });
        return hash;
    };
    return UnionType;
}(NamedType));
exports.UnionType = UnionType;
function typesEqual(t1, t2) {
    if (t1 === t2)
        return true;
    var queueOrResult = t1.expandForEquality(t2);
    if (typeof queueOrResult === "boolean")
        return queueOrResult;
    var queue = queueOrResult;
    var alreadySeenByHash = {};
    function alreadySeen(types) {
        var hash = types[0].hashCode().toString();
        var pairs = alreadySeenByHash[hash];
        if (pairs) {
            for (var _i = 0, pairs_1 = pairs; _i < pairs_1.length; _i++) {
                var _a = pairs_1[_i], o1 = _a[0], o2 = _a[1];
                if (o1 === types[0] && o2 === types[1])
                    return true;
            }
        }
        else {
            alreadySeenByHash[hash] = pairs = [];
        }
        pairs.push(types);
        return false;
    }
    for (;;) {
        var maybePair = queue.pop();
        if (!maybePair)
            return true;
        t1 = maybePair[0], t2 = maybePair[1];
        if (t1 === t2)
            continue;
        if (alreadySeen(maybePair))
            continue;
        queueOrResult = t1.expandForEquality(t2);
        if (typeof queueOrResult === "boolean") {
            if (!queueOrResult)
                return false;
            continue;
        }
        for (var _i = 0, queueOrResult_1 = queueOrResult; _i < queueOrResult_1.length; _i++) {
            var p = queueOrResult_1[_i];
            queue.push(p);
        }
    }
}
function removeNullFromUnion(t) {
    var nullType = t.findMember("null");
    if (!nullType) {
        return [null, t.members];
    }
    return [nullType, t.members.filterNot(isNull).toOrderedSet()];
}
exports.removeNullFromUnion = removeNullFromUnion;
function nullableFromUnion(t) {
    var _a = removeNullFromUnion(t), hasNull = _a[0], nonNulls = _a[1];
    if (!hasNull)
        return null;
    if (nonNulls.size !== 1)
        return null;
    return Support_1.defined(nonNulls.first());
}
exports.nullableFromUnion = nullableFromUnion;
function makeNullable(t, typeNames, areNamesInferred) {
    if (t.kind === "null") {
        return t;
    }
    if (!(t instanceof UnionType)) {
        return new UnionType(typeNames, areNamesInferred, immutable_1.OrderedSet([t, new PrimitiveType("null")]));
    }
    var _a = removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
    if (maybeNull)
        return t;
    return new UnionType(typeNames, areNamesInferred, nonNulls.add(new PrimitiveType("null")));
}
exports.makeNullable = makeNullable;
function removeNull(t) {
    if (!(t instanceof UnionType)) {
        return t;
    }
    var _a = removeNullFromUnion(t), _ = _a[0], nonNulls = _a[1];
    var first = nonNulls.first();
    if (first) {
        if (nonNulls.size === 1)
            return first;
        return new UnionType(t.names, t.areNamesInferred, nonNulls);
    }
    return Support_1.panic("Trying to remove null results in empty union.");
}
exports.removeNull = removeNull;
// FIXME: The outer OrderedSet should be some Collection, but I can't figure out
// which one.  Collection.Indexed doesn't work with OrderedSet, which is unfortunate.
function orderedSetUnion(sets) {
    var setArray = sets.toArray();
    if (setArray.length === 0)
        return immutable_1.OrderedSet();
    if (setArray.length === 1)
        return setArray[0];
    return (_a = setArray[0]).union.apply(_a, setArray.slice(1));
    var _a;
}
function filterTypes(predicate, graph, childrenOfType) {
    var seen = immutable_1.Set();
    var types = immutable_1.List();
    function addFromType(t) {
        if (seen.has(t))
            return;
        seen = seen.add(t);
        var children = childrenOfType ? childrenOfType(t) : t.children;
        children.forEach(addFromType);
        if (predicate(t)) {
            types = types.push(t);
        }
    }
    graph.forEach(addFromType);
    return types.reverse().toOrderedSet();
}
exports.filterTypes = filterTypes;
function allNamedTypes(graph, childrenOfType) {
    return filterTypes(function (t) { return t.isNamedType(); }, graph, childrenOfType);
}
exports.allNamedTypes = allNamedTypes;
function separateNamedTypes(types) {
    var classes = types.filter(function (t) { return t instanceof ClassType; }).toOrderedSet();
    var enums = types.filter(function (t) { return t instanceof EnumType; }).toOrderedSet();
    var unions = types.filter(function (t) { return t instanceof UnionType; }).toOrderedSet();
    return { classes: classes, enums: enums, unions: unions };
}
exports.separateNamedTypes = separateNamedTypes;
function allNamedTypesSeparated(graph, childrenOfType) {
    var types = allNamedTypes(graph, childrenOfType);
    return separateNamedTypes(types);
}
exports.allNamedTypesSeparated = allNamedTypesSeparated;
function matchType(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType) {
    if (t instanceof PrimitiveType) {
        var f = {
            any: anyType,
            null: nullType,
            bool: boolType,
            integer: integerType,
            double: doubleType,
            string: stringType
        }[t.kind];
        if (f)
            return f(t);
        return Support_1.panic("Unknown PrimitiveType: " + t.kind);
    }
    else if (t instanceof ArrayType)
        return arrayType(t);
    else if (t instanceof ClassType)
        return classType(t);
    else if (t instanceof MapType)
        return mapType(t);
    else if (t instanceof EnumType)
        return enumType(t);
    else if (t instanceof UnionType)
        return unionType(t);
    return Support_1.panic("Unknown Type");
}
exports.matchType = matchType;
