import { Map, OrderedSet } from "immutable";
import { PrimitiveTypeKind, Type, PrimitiveType, EnumType, MapType, ArrayType, ClassType, UnionType } from "./Type";
export declare class TypeBuilder {
    private _primitiveTypes;
    private _mapTypes;
    private _arrayTypes;
    private _enumTypes;
    private _classTypes;
    private _unionTypes;
    getPrimitiveType: (kind: PrimitiveTypeKind) => PrimitiveType;
    getEnumType: (name: string, isInferred: boolean, cases: OrderedSet<string>) => EnumType;
    getMapType: (values: Type) => MapType;
    getArrayType: (items: Type) => ArrayType;
    getClassType: (name: string, isInferred: boolean, properties: Map<string, Type>) => ClassType;
    getUniqueClassType: (name: string, isInferred: boolean, properties?: Map<string, Type> | undefined) => ClassType;
    getUnionType: (name: string, isInferred: boolean, members: OrderedSet<Type>) => UnionType;
    getUniqueUnionType: (name: string, isInferred: boolean, members: OrderedSet<Type>) => UnionType;
}
export declare abstract class UnionBuilder<TArray, TClass, TMap> {
    protected readonly typeBuilder: TypeBuilder;
    protected readonly typeName: string;
    protected readonly isInferred: boolean;
    private _haveAny;
    private _haveNull;
    private _haveBool;
    private _haveInteger;
    private _haveDouble;
    private _haveString;
    private readonly _arrays;
    private readonly _maps;
    private readonly _classes;
    private _enumCaseMap;
    private _enumCases;
    constructor(typeBuilder: TypeBuilder, typeName: string, isInferred: boolean);
    readonly haveString: boolean;
    addAny: () => void;
    addNull: () => void;
    addBool: () => void;
    addInteger: () => void;
    addDouble: () => void;
    addString: () => void;
    addArray: (t: TArray) => void;
    addClass: (t: TClass) => void;
    addMap: (t: TMap) => void;
    addEnumCase: (s: string) => void;
    protected abstract makeEnum(cases: string[]): Type | null;
    protected abstract makeClass(classes: TClass[], maps: TMap[]): Type;
    protected abstract makeArray(arrays: TArray[]): Type;
    buildUnion: (unique: boolean) => Type;
}
