#!/usr/bin/env node
"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var process = require("process");
var getStream = require("get-stream");
var _ = require("lodash");
var targetLanguages = require("./Language/All");
var Annotation_1 = require("./Annotation");
var CompressedJSON_1 = require("./CompressedJSON");
var URLGrammar_1 = require("./URLGrammar");
var commandLineArgs = require("command-line-args");
var getUsage = require("command-line-usage");
var fetch = require("node-fetch");
var chalk = require("chalk");
var langs = targetLanguages.all.map(function (r) { return _.minBy(r.names, function (s) { return s.length; }); }).join("|");
var langDisplayNames = targetLanguages.all.map(function (r) { return r.displayName; }).join(", ");
var optionDefinitions = [
    {
        name: "out",
        alias: "o",
        type: String,
        typeLabel: "FILE",
        description: "The output file. Determines --lang and --top-level."
    },
    {
        name: "top-level",
        alias: "t",
        type: String,
        typeLabel: "NAME",
        description: "The name for the top level type."
    },
    {
        name: "lang",
        alias: "l",
        type: String,
        typeLabel: langs,
        description: "The target language."
    },
    {
        name: "src-lang",
        alias: "s",
        type: String,
        defaultValue: "json",
        typeLabel: "json|schema",
        description: "The source language (default is json)."
    },
    {
        name: "src",
        type: String,
        multiple: true,
        defaultOption: true,
        typeLabel: "FILE|URL|DIRECTORY",
        description: "The file, url, or data directory to type."
    },
    {
        name: "src-urls",
        type: String,
        typeLabel: "FILE",
        description: "Tracery grammar describing URLs to crawl."
    },
    {
        name: "no-combine-classes",
        type: Boolean,
        description: "Don't combine similar classes."
    },
    {
        name: "no-maps",
        type: Boolean,
        description: "Don't infer maps, always use classes."
    },
    {
        name: "no-enums",
        type: Boolean,
        description: "Don't infer enums, always use strings."
    },
    {
        name: "no-render",
        type: Boolean,
        description: "Don't render output."
    },
    {
        name: "quiet",
        type: Boolean,
        description: "Don't show issues in the generated code."
    },
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Get some help."
    }
];
var sectionsBeforeRenderers = [
    {
        header: "Synopsis",
        content: "$ quicktype [[bold]{--lang} " + langs + "] FILE|URL ..."
    },
    {
        header: "Description",
        content: "Given JSON sample data, quicktype outputs code for working with that data in " + langDisplayNames + "."
    },
    {
        header: "Options",
        optionList: optionDefinitions,
        hide: ["no-render"]
    }
];
var sectionsAfterRenderers = [
    {
        header: "Examples",
        content: [
            chalk.dim("Generate C# to parse a Bitcoin API"),
            "$ quicktype -o LatestBlock.cs https://blockchain.info/latestblock",
            "",
            chalk.dim("Generate Go code from a directory of samples containing:"),
            chalk.dim("  - Foo.json\n  + Bar\n    - bar-sample-1.json\n    - bar-sample-2.json\n  - Baz.url"),
            "$ quicktype -l go samples",
            "",
            chalk.dim("Generate JSON Schema, then TypeScript"),
            "$ quicktype -o schema.json https://blockchain.info/latestblock",
            "$ quicktype -o bitcoin.ts --src-lang schema schema.json"
        ]
    },
    {
        content: "Learn more at [bold]{quicktype.io}"
    }
];
function getTargetLanguage(name) {
    var language = targetLanguages.languageNamed(name);
    if (language) {
        return language;
    }
    console.error("'" + name + "' is not yet supported as an output language.");
    return process.exit(1);
}
function usage() {
    var rendererSections = [];
    _.forEach(targetLanguages.all, function (language) {
        var definitions = language.optionDefinitions;
        if (definitions.length === 0)
            return;
        rendererSections.push({
            header: "Options for " + language.displayName,
            optionList: definitions
        });
    });
    var sections = _.concat(sectionsBeforeRenderers, rendererSections, sectionsAfterRenderers);
    console.log(getUsage(sections));
}
var Run = /** @class */ (function () {
    function Run(argv) {
        var _this = this;
        this.getOptionDefinitions = function (opts) {
            return getTargetLanguage(opts.lang).optionDefinitions;
        };
        this.renderSamplesOrSchemas = function () {
            var targetLanguage = getTargetLanguage(_this._options.lang);
            var topLevels;
            if (_this.isInputJSONSchema) {
                var names = Object.getOwnPropertyNames(_this._allSamples.schemas);
                topLevels = names.map(function (name) { return ({ name: name, schema: _this._allSamples.schemas[name] }); });
            }
            else {
                var names = Object.getOwnPropertyNames(_this._allSamples.samples);
                topLevels = names.map(function (name) { return ({ name: name, samples: _this._allSamples.samples[name] }); });
            }
            var config = {
                language: targetLanguage.names[0],
                isInputJSONSchema: _this.isInputJSONSchema,
                topLevels: topLevels,
                compressedJSON: _this._compressedJSON,
                inferMaps: !_this._options.noMaps,
                inferEnums: !_this._options.noEnums,
                combineClasses: !_this._options.noCombineClasses,
                doRender: !_this._options.noRender,
                rendererOptions: _this._options.rendererOptions
            };
            try {
                return targetLanguage.transformAndRenderConfig(config);
            }
            catch (e) {
                console.error(e);
                return process.exit(1);
            }
        };
        this.splitAndWriteJava = function (dir, str) {
            var lines = str.split("\n");
            var filename = null;
            var currentFileContents = "";
            var writeFile = function () {
                if (filename != null) {
                    fs.writeFileSync(path.join(dir, filename), currentFileContents);
                }
                filename = null;
                currentFileContents = "";
            };
            var i = 0;
            while (i < lines.length) {
                var line = lines[i];
                i += 1;
                var results = line.match("^// (.+\\.java)$");
                if (results == null) {
                    currentFileContents += line + "\n";
                }
                else {
                    writeFile();
                    filename = results[1];
                    while (lines[i] === "")
                        i++;
                }
            }
            writeFile();
        };
        this.render = function () {
            var _a = _this.renderSamplesOrSchemas(), lines = _a.lines, annotations = _a.annotations;
            var output = lines.join("\n");
            if (_this._options.out) {
                if (_this._options.lang === "java") {
                    _this.splitAndWriteJava(path.dirname(_this._options.out), output);
                }
                else {
                    fs.writeFileSync(_this._options.out, output);
                }
            }
            else {
                process.stdout.write(output);
            }
            if (_this._options.quiet) {
                return;
            }
            annotations.forEach(function (sa) {
                var annotation = sa.annotation;
                if (!(annotation instanceof Annotation_1.IssueAnnotationData))
                    return;
                var lineNumber = sa.span.start.line;
                var humanLineNumber = lineNumber + 1;
                console.error("\nIssue in line " + humanLineNumber + ": " + annotation.message);
                console.error(humanLineNumber + ": " + lines[lineNumber]);
            });
        };
        this.readSampleFromStream = function (name, readStream) { return __awaiter(_this, void 0, void 0, function () {
            var input, _a, _b;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!this.needCompressedJSONInput) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._compressedJSON.readFromStream(readStream)];
                    case 1:
                        input = _c.sent();
                        return [3 /*break*/, 4];
                    case 2:
                        _b = (_a = JSON).parse;
                        return [4 /*yield*/, getStream(readStream)];
                    case 3:
                        input = _b.apply(_a, [_c.sent()]);
                        _c.label = 4;
                    case 4:
                        if (this.isInputJSONSchema) {
                            if (Object.prototype.hasOwnProperty.call(this._allSamples.schemas, name)) {
                                console.error("Error: More than one schema given for top-level " + name + ".");
                                return [2 /*return*/, process.exit(1)];
                            }
                            this._allSamples.schemas[name] = input;
                        }
                        else {
                            if (!Object.prototype.hasOwnProperty.call(this._allSamples.samples, name)) {
                                this._allSamples.samples[name] = [];
                            }
                            this._allSamples.samples[name].push(input);
                        }
                        return [2 /*return*/];
                }
            });
        }); };
        this.readSampleFromFileOrUrl = function (name, fileOrUrl) { return __awaiter(_this, void 0, void 0, function () {
            var res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!fs.existsSync(fileOrUrl)) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.readSampleFromStream(name, fs.createReadStream(fileOrUrl))];
                    case 1:
                        _a.sent();
                        return [3 /*break*/, 5];
                    case 2: return [4 /*yield*/, fetch(fileOrUrl)];
                    case 3:
                        res = _a.sent();
                        return [4 /*yield*/, this.readSampleFromStream(name, res.body)];
                    case 4:
                        _a.sent();
                        _a.label = 5;
                    case 5: return [2 /*return*/];
                }
            });
        }); };
        this.readSampleFromFileOrUrlArray = function (name, filesOrUrls) { return __awaiter(_this, void 0, void 0, function () {
            var _i, filesOrUrls_1, fileOrUrl;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _i = 0, filesOrUrls_1 = filesOrUrls;
                        _a.label = 1;
                    case 1:
                        if (!(_i < filesOrUrls_1.length)) return [3 /*break*/, 4];
                        fileOrUrl = filesOrUrls_1[_i];
                        return [4 /*yield*/, this.readSampleFromFileOrUrl(name, fileOrUrl)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        _i++;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        }); };
        this.readNamedSamplesFromDirectory = function (dataDir) { return __awaiter(_this, void 0, void 0, function () {
            var _this = this;
            var readFilesOrURLsInDirectory, contents, directories, _i, directories_1, dir, sampleName;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        readFilesOrURLsInDirectory = function (d, sampleName) { return __awaiter(_this, void 0, void 0, function () {
                            var files, _i, files_1, file, name, fileOrUrl;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        files = fs
                                            .readdirSync(d)
                                            .map(function (x) { return path.join(d, x); })
                                            .filter(function (x) { return fs.lstatSync(x).isFile(); });
                                        _i = 0, files_1 = files;
                                        _a.label = 1;
                                    case 1:
                                        if (!(_i < files_1.length)) return [3 /*break*/, 4];
                                        file = files_1[_i];
                                        if (sampleName === undefined) {
                                            name = path.basename(file);
                                            sampleName = name.substr(0, name.lastIndexOf("."));
                                        }
                                        fileOrUrl = file;
                                        // If file is a URL string, download it
                                        if (_.endsWith(file, ".url")) {
                                            fileOrUrl = fs.readFileSync(file, "utf8").trim();
                                        }
                                        return [4 /*yield*/, this.readSampleFromFileOrUrl(sampleName, fileOrUrl)];
                                    case 2:
                                        _a.sent();
                                        _a.label = 3;
                                    case 3:
                                        _i++;
                                        return [3 /*break*/, 1];
                                    case 4: return [2 /*return*/];
                                }
                            });
                        }); };
                        contents = fs.readdirSync(dataDir).map(function (x) { return path.join(dataDir, x); });
                        directories = contents.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                        return [4 /*yield*/, readFilesOrURLsInDirectory(dataDir)];
                    case 1:
                        _a.sent();
                        _i = 0, directories_1 = directories;
                        _a.label = 2;
                    case 2:
                        if (!(_i < directories_1.length)) return [3 /*break*/, 5];
                        dir = directories_1[_i];
                        sampleName = path.basename(dir);
                        return [4 /*yield*/, readFilesOrURLsInDirectory(dir, sampleName)];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/];
                }
            });
        }); };
        this.main = function () { return __awaiter(_this, void 0, void 0, function () {
            var json, jsonMap, _i, _a, key, exists, directories_2, _b, directories_3, dataDir, filesOrUrls;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!this._options.help) return [3 /*break*/, 1];
                        usage();
                        return [2 /*return*/];
                    case 1:
                        if (!this._options.srcUrls) return [3 /*break*/, 6];
                        json = JSON.parse(fs.readFileSync(this._options.srcUrls, "utf8"));
                        jsonMap = URLGrammar_1.urlsFromURLGrammar(json);
                        _i = 0, _a = Object.keys(jsonMap);
                        _c.label = 2;
                    case 2:
                        if (!(_i < _a.length)) return [3 /*break*/, 5];
                        key = _a[_i];
                        return [4 /*yield*/, this.readSampleFromFileOrUrlArray(key, jsonMap[key])];
                    case 3:
                        _c.sent();
                        _c.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [3 /*break*/, 14];
                    case 6:
                        if (!(this._options.src.length === 0)) return [3 /*break*/, 8];
                        // FIXME: Why do we have to convert to any here?
                        return [4 /*yield*/, this.readSampleFromStream(this._options.topLevel, process.stdin)];
                    case 7:
                        // FIXME: Why do we have to convert to any here?
                        _c.sent();
                        return [3 /*break*/, 14];
                    case 8:
                        exists = this._options.src.filter(fs.existsSync);
                        directories_2 = exists.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                        _b = 0, directories_3 = directories_2;
                        _c.label = 9;
                    case 9:
                        if (!(_b < directories_3.length)) return [3 /*break*/, 12];
                        dataDir = directories_3[_b];
                        return [4 /*yield*/, this.readNamedSamplesFromDirectory(dataDir)];
                    case 10:
                        _c.sent();
                        _c.label = 11;
                    case 11:
                        _b++;
                        return [3 /*break*/, 9];
                    case 12:
                        filesOrUrls = this._options.src.filter(function (x) { return !_.includes(directories_2, x); });
                        if (!!_.isEmpty(filesOrUrls)) return [3 /*break*/, 14];
                        return [4 /*yield*/, this.readSampleFromFileOrUrlArray(this._options.topLevel, filesOrUrls)];
                    case 13:
                        _c.sent();
                        _c.label = 14;
                    case 14:
                        this.render();
                        return [2 /*return*/];
                }
            });
        }); };
        // Parse the options in argv and split them into global options and renderer options,
        // according to each option definition's `renderer` field.  If `partial` is false this
        // will throw if it encounters an unknown option.
        this.parseOptions = function (definitions, argv, partial) {
            var opts = commandLineArgs(definitions, {
                argv: argv,
                partial: partial
            });
            var options = { rendererOptions: {} };
            definitions.forEach(function (o) {
                if (!(o.name in opts))
                    return;
                var v = opts[o.name];
                if (o.renderer)
                    options.rendererOptions[o.name] = v;
                else {
                    var k = _.lowerFirst(o.name
                        .split("-")
                        .map(_.upperFirst)
                        .join(""));
                    options[k] = v;
                }
            });
            return _this.inferOptions(options);
        };
        this.inferOptions = function (opts) {
            return __assign({ src: opts.src || [], srcLang: opts.srcLang || "json", lang: opts.lang || _this.inferLang(opts), topLevel: opts.topLevel || _this.inferTopLevel(opts), noMaps: !!opts.noMaps, noEnums: !!opts.noEnums, noCombineClasses: !!opts.noCombineClasses, noRender: !!opts.noRender, help: !!opts.help, quiet: !!opts.quiet }, opts);
        };
        this.inferLang = function (options) {
            // Output file extension determines the language if language is undefined
            if (options.out) {
                var extension = path.extname(options.out);
                if (extension === "") {
                    console.error("Please specify a language (--lang) or an output file extension.");
                    process.exit(1);
                }
                return extension.substr(1);
            }
            return "go";
        };
        this.inferTopLevel = function (options) {
            // Output file name determines the top-level if undefined
            if (options.out) {
                var extension = path.extname(options.out);
                var without = path.basename(options.out).replace(extension, "");
                return without;
            }
            // Source determines the top-level if undefined
            if (options.src && options.src.length === 1) {
                var src = options.src[0];
                var extension = path.extname(src);
                var without = path.basename(src).replace(extension, "");
                return without;
            }
            return "TopLevel";
        };
        if (_.isArray(argv)) {
            // We can only fully parse the options once we know which renderer is selected,
            // because there are renderer-specific options.  But we only know which renderer
            // is selected after we've parsed the options.  Hence, we parse the options
            // twice.  This is the first parse to get the renderer:
            var incompleteOptions = this.parseOptions(optionDefinitions, argv, true);
            var rendererOptionDefinitions = this.getOptionDefinitions(incompleteOptions);
            // Use the global options as well as the renderer options from now on:
            var allOptionDefinitions = _.concat(optionDefinitions, rendererOptionDefinitions);
            try {
                // This is the parse that counts:
                this._options = this.parseOptions(allOptionDefinitions, argv, false);
            }
            catch (error) {
                if (error.name === "UNKNOWN_OPTION") {
                    console.error("Error: Unknown option");
                    usage();
                    process.exit(1);
                }
                throw error;
            }
        }
        else {
            this._options = this.inferOptions(argv);
        }
        this._compressedJSON = new CompressedJSON_1.CompressedJSON();
        this._allSamples = { samples: {}, schemas: {} };
    }
    Object.defineProperty(Run.prototype, "isInputJSONSchema", {
        get: function () {
            return this._options.srcLang === "schema";
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Run.prototype, "needCompressedJSONInput", {
        get: function () {
            if (this.isInputJSONSchema) {
                return false;
            }
            var lang = getTargetLanguage(this._options.lang);
            return lang.needsCompressedJSONInput(this._options.rendererOptions);
        },
        enumerable: true,
        configurable: true
    });
    return Run;
}());
function main(args) {
    return __awaiter(this, void 0, void 0, function () {
        var run;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!(_.isArray(args) && args.length === 0)) return [3 /*break*/, 1];
                    usage();
                    return [3 /*break*/, 3];
                case 1:
                    run = new Run(args);
                    return [4 /*yield*/, run.main()];
                case 2:
                    _a.sent();
                    _a.label = 3;
                case 3: return [2 /*return*/];
            }
        });
    });
}
exports.main = main;
if (require.main === module) {
    main(process.argv.slice(2)).catch(function (reason) {
        console.error(reason);
        process.exit(1);
    });
}
